from argparse import ArgumentParser, RawTextHelpFormatter
from asyncio import iscoroutinefunction
from typing import Any, Awaitable, Callable, Coroutine, Dict, List, NamedTuple, Optional, Tuple, Union

__all__ = (
    # commander_app
    'Command',
    'command',
    'CommandHandler',
    'Application',
)

CommandHandler = Callable[[Dict[str, Any]], Union[Callable[[Any], int], Coroutine[Any, Any, int]]]


class Command(NamedTuple):
    name: str
    handler: CommandHandler
    positionals: List[Tuple[str, Dict[str, Any]]]
    optionals: List[Tuple[str, Dict[str, Any]]]


def command(
    name: str,
    handler: CommandHandler,
    positionals: Optional[List[Tuple[str, Dict[str, Any]]]] = None,
    optionals: Optional[List[Tuple[str, Dict[str, Any]]]] = None
) -> Command:
    return Command(name=name, handler=handler, positionals=positionals or [], optionals=optionals or [])


AppSignal = Callable[['Application'], Union[None, Awaitable[None]]]


class Application:
    _parser: ArgumentParser
    _parsers: Dict[str, ArgumentParser]
    _commands: Dict[str, Command]
    _exit_code: int
    _on_startup: List[AppSignal]
    _on_shutdown: List[AppSignal]
    _on_cleanup: List[AppSignal]

    def __init__(
        self,
        commands: Optional[List[Command]] = None,
        *,
        name: str = 'aiocli.commander',
        description: str = '',
        version: str = 'unknown',
        on_startup: Optional[List[AppSignal]] = None,
        on_shutdown: Optional[List[AppSignal]] = None,
        on_cleanup: Optional[List[AppSignal]] = None,
    ) -> None:
        self._parser = ArgumentParser(
            description=description,
            prog=name,
            formatter_class=RawTextHelpFormatter,
        )
        self._parser.add_argument('--version', action='version', version=version)
        self._parsers = {}
        self._commands = {}
        self.add_commands(commands or [])
        self._exit_code = 0
        self._on_startup = on_startup if on_startup else []
        self._on_shutdown = on_shutdown if on_shutdown else []
        self._on_cleanup = on_cleanup if on_cleanup else []

    async def __call__(self, args: List[str]) -> int:
        command_name = args[0] if len(args) > 0 else None
        is_optional_argument = command_name and command_name[0] == '-'
        if command_name not in self._parsers and not is_optional_argument:
            self._parser.print_help()
            return 0
        if is_optional_argument:
            self._parser.parse_args()
            return 0
        namespace = self._parsers[command_name].parse_args(args[1:])
        handler = self._commands[command_name].handler
        self._exit_code = await handler(vars(namespace)) if iscoroutinefunction(handler) else handler(vars(namespace))
        return self._exit_code

    def include_router(self, router: 'Application') -> None:
        for name, cmd in router._commands.items():
            self._commands[name] = cmd
            self._parsers[name] = router._parsers[name]
            self._update_parser_description(name)
        for on_startup in router._on_startup:
            self._on_startup.append(on_startup)
        for on_shutdown in router._on_shutdown:
            self._on_shutdown.append(on_shutdown)
        for on_cleanup in router._on_cleanup:
            self._on_cleanup.append(on_cleanup)

    def command(
        self,
        name: str,
        *,
        positionals: Optional[List[Tuple[str, Dict[str, Any]]]] = None,
        optionals: Optional[List[Tuple[str, Dict[str, Any]]]] = None,
    ) -> Callable[[CommandHandler], CommandHandler]:
        def decorator(handler: CommandHandler) -> CommandHandler:
            self._add_command(Command(
                name=name,
                handler=handler,
                positionals=positionals or [],
                optionals=optionals or [],
            ))
            return handler

        return decorator

    def add_commands(self, commands: List[Command]) -> None:
        for cmd in commands:
            self._add_command(cmd)

    def get_command(self, name: str) -> Optional[Command]:
        return self._commands.get(name, None)

    @property
    def parser(self) -> ArgumentParser:
        return self._parser

    @property
    def exit_code(self) -> int:
        return self._exit_code

    @property
    def on_startup(self) -> List[AppSignal]:
        return self._on_startup

    async def startup(self) -> None:
        await self._execute_app_signals(self.on_startup)

    @property
    def on_shutdown(self) -> List[AppSignal]:
        return self._on_shutdown

    async def shutdown(self) -> None:
        await self._execute_app_signals(self._on_shutdown)

    @property
    def on_cleanup(self) -> List[AppSignal]:
        return self._on_cleanup

    async def cleanup(self) -> None:
        await self._execute_app_signals(self._on_cleanup)

    def _add_command(self, cmd: Command) -> None:
        self._commands[cmd.name] = cmd
        parser = ArgumentParser(prog=cmd.name)
        _ = [parser.add_argument(arg[0], **arg[1]) for arg in cmd.positionals + cmd.optionals]
        self._parsers[cmd.name] = parser
        self._update_parser_description(cmd.name)

    async def _execute_app_signals(self, app_signals: List[AppSignal]) -> None:
        for app_signal in app_signals:
            await app_signal(self) if iscoroutinefunction(app_signal) else app_signal(self)

    def _update_parser_description(self, name: str) -> None:
        self._parser.description = '{0}{1}\n'.format(self._parser.description, name)



import signal
import sys
from asyncio import get_event_loop, gather, all_tasks, iscoroutine
from asyncio.events import AbstractEventLoop
from typing import Awaitable, Optional, Union, cast, List

from aiocli.commander_app import Application, Command, command

__all__ = (
    # commander_app
    'Application',
    'Command',
    'command',
    # commander
    'run_app',
)


class GracefulExit(SystemExit):
    code = 1


def _raise_graceful_exit() -> None:
    raise GracefulExit()


def _cancel_all_tasks(loop: AbstractEventLoop) -> None:
    to_cancel = all_tasks(loop)
    if not to_cancel:
        return
    for task in to_cancel:
        task.cancel()
    loop.run_until_complete(gather(*to_cancel, loop=loop, return_exceptions=True))
    for task in to_cancel:
        if task.cancelled():
            continue
        if task.exception() is not None:
            loop.call_exception_handler({
                'message': 'unhandled exception during asyncio.run() shutdown',
                'exception': task.exception(),
                'task': task,
            })


class AppRunner:
    __slots__ = ('_app', '_loop', '_handle_signals', '_exit_code')

    def __init__(
            self,
            app: Application,
            *,
            loop: Optional[AbstractEventLoop] = None,
            handle_signals: bool = False,
            exit_code: bool = False
    ) -> None:
        self._app = app
        self._loop = loop or get_event_loop()
        self._handle_signals = handle_signals
        self._exit_code = exit_code

    @property
    def app(self) -> Application:
        return self._app

    async def setup(self) -> None:
        if self._handle_signals:
            try:
                self._loop.add_signal_handler(signal.SIGINT, _raise_graceful_exit)
                self._loop.add_signal_handler(signal.SIGTERM, _raise_graceful_exit)
            except NotImplementedError:  # pragma: no cover
                # add_signal_handler is not implemented on Windows
                pass
        await self.startup()

    async def startup(self) -> None:
        await self._app.startup()

    async def shutdown(self) -> None:
        await self._app.shutdown()

    async def cleanup(self) -> None:
        await self.shutdown()
        if self._handle_signals:
            try:
                self._loop.remove_signal_handler(signal.SIGINT)
                self._loop.remove_signal_handler(signal.SIGTERM)
            except NotImplementedError:  # pragma: no cover
                # remove_signal_handler is not implemented on Windows
                pass
        await self._app.cleanup()
        if self._exit_code:
            self._app.parser.exit(status=self._app.exit_code)


async def _run_app(
        app: Union[Application, Awaitable[Application]],
        *,
        loop: AbstractEventLoop,
        handle_signals: bool = True,
        argv: Optional[List[str]] = None,
        exit_code: bool = True
) -> None:
    app = cast(Application, await app if iscoroutine(app) else app)  # type: ignore
    runner = AppRunner(app, loop=loop, handle_signals=handle_signals, exit_code=exit_code)
    await runner.setup()
    try:
        await app(argv or sys.argv[1:])
    finally:
        await runner.cleanup()


def run_app(
        app: Union[Application, Awaitable[Application]],
        *,
        loop: Optional[AbstractEventLoop] = None,
        handle_signals: bool = True,
        argv: Optional[List[str]] = None,
        exit_code: bool = True
) -> None:
    loop = loop or get_event_loop()
    try:
        loop.run_until_complete(_run_app(app, loop=loop, handle_signals=handle_signals, argv=argv, exit_code=exit_code))
    except (GracefulExit, KeyboardInterrupt):  # pragma: no cover
        pass
    finally:
        _cancel_all_tasks(loop)
        loop.run_until_complete(loop.shutdown_asyncgens())
        loop.close()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiocli']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'aiocli',
    'version': '1.2.1',
    'description': 'Simple and lightweight async console runner.',
    'long_description': "# Async cli client/commander framework\n\naiocli is a Python library for simple and lightweight async console runner.\n\nFull compatibility with argparse module and highly inspired by aiohttp module.\n\n## Installation\n\nUse the package manager [pip](https://pip.pypa.io/en/stable/) to install aiocli.\n\n```bash\npip install aiocli\n```\n\n## Usage\n\n```python\n# examples/commander.py\nfrom aiocli import commander\n\napp = commander.Application()\n\n@app.command(name='greet:to', positionals=[('--name', {'default': 'World!'})])\nasync def handle(args: dict) -> int:\n    print('Hello ' + args.get('name'))\n    return 0\n\nif __name__ == '__main__':\n    commander.run_app(app)\n```\n\n## Requirements\n\n- Python >= 3.6\n\n## Contributing\nPull requests are welcome. For major changes, please open an issue first to discuss what you would like to change.\n\nPlease make sure to update tests as appropriate.\n\n## License\n[MIT](https://github.com/ticdenis/python-aiocli/blob/master/LICENSE)\n",
    'author': 'ticdenis',
    'author_email': 'denisnavarroalcaide@outlook.es',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ticdenis/python-aiocli',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
